﻿using System;
using System.Collections.Generic;
using System.Text;
using SimioAPI;
using SimioAPI.Extensions;

namespace CustomSimioStep
{
    class FileElementDefinition : IElementDefinition
    {
        #region IElementDefinition Members

        /// <summary>
        /// Property returning the full name for this type of element. The name should contain no spaces. 
        /// </summary>
        public string Name
        {
            get { return "File"; }
        }

        /// <summary>
        /// Property returning a short description of what the element does.  
        /// </summary>
        public string Description
        {
            get { return "The File element may be used in conjunction with the user defined Read and Write steps to read and write to an external file."; }
        }

        /// <summary>
        /// Property returning an icon to display for the element in the UI. 
        /// </summary>
        public System.Drawing.Image Icon
        {
            get { return null; }
        }

        /// <summary>
        /// Property returning a unique static GUID for the element.  
        /// </summary>
        public Guid UniqueID
        {
            get { return MY_ID; }
        }
        // We need to use this ID in the element reference property of the Read/Write steps, so we make it public
        public static readonly Guid MY_ID = new Guid("{517F0463-A69A-49c0-AA8F-830A8C3A9FF9}");

        /// <summary>
        /// Method called that defines the property, state, and event schema for the element.
        /// </summary>
        public void DefineSchema(IElementSchema schema)
        {
            IPropertyDefinition pd = schema.PropertyDefinitions.AddStringProperty("FilePath", String.Empty);
            pd.Description = "The name of the actual file that is being read from or written to.";
        }

        /// <summary>
        /// Method called to add a new instance of this element type to a model. 
        /// Returns an instance of the class implementing the IElement interface.
        /// </summary>
        public IElement CreateElement(IElementData data)
        {
            return new FileElement(data);
        }

        #endregion
    }

    class FileElement : IElement
    {
        IElementData _data;
        string _writerFileName;
        string _readerFileName;
        public FileElement(IElementData data)
        {
            _data = data;
            IPropertyReader fileNameProp = _data.Properties.GetProperty("FilePath");
            
            // Cache the names of the files to open for reading or writing
            string fileName = fileNameProp.GetStringValue(_data.ExecutionContext);
            if (String.IsNullOrEmpty(fileName) == false)
            {
                string fileRoot = null;
                string fileDirectoryName = null;
                string fileNameWithoutExtension = null;
                string fileExtension = null;

                try
                {
                    fileRoot = System.IO.Path.GetPathRoot(fileName);
                    fileDirectoryName = System.IO.Path.GetDirectoryName(fileName);
                    fileNameWithoutExtension = System.IO.Path.GetFileNameWithoutExtension(fileName);
                    fileExtension = System.IO.Path.GetExtension(fileName);
                }
                catch (ArgumentException e)
                {
                    data.ExecutionContext.ExecutionInformation.ReportError(String.Format("Failed to create runtime file element. Message: {0}", e.Message));
                }

                string simioProjectFolder = _data.ExecutionContext.ExecutionInformation.ProjectFolder;
                string simioExperimentName = _data.ExecutionContext.ExecutionInformation.ExperimentName;
                string simioScenarioName = _data.ExecutionContext.ExecutionInformation.ScenarioName;
                string simioReplicationNumber = _data.ExecutionContext.ExecutionInformation.ReplicationNumber.ToString();

                if (String.IsNullOrEmpty(fileDirectoryName) || String.IsNullOrEmpty(fileRoot))
                {
                    fileDirectoryName = simioProjectFolder;
                    fileName = fileDirectoryName + "\\" + fileName;
                }

                _readerFileName = fileName;

                if (String.IsNullOrEmpty(simioExperimentName))
                    _writerFileName = fileName;
                else
                    //_writerFileName = fileDirectoryName + "\\" + fileNameWithoutExtension + "_" + simioExperimentName + "_" + simioScenarioName + "_Rep" + simioReplicationNumber + fileExtension;
                    _writerFileName = fileName + "_" + simioScenarioName + "_Rep" + simioReplicationNumber + fileExtension;
            }
        }

        System.IO.TextWriter _writer;
        public System.IO.TextWriter Writer
        {
            get
            {
                // We can't read and write at the same time
                if (_reader != null)
                {
                    _data.ExecutionContext.ExecutionInformation.ReportError(String.Format("Trying to write to {0}, which is already open for reading.", _writerFileName ?? "[No file specified]"));
                    return null;
                }

                // If we don't already have a writer, create one
                try
                {
                    if (String.IsNullOrEmpty(_writerFileName))
                        ReportFileOpenError("[No file specified]", "writing", "[None]");
                    else if (_writer == null)
                        _writer = new System.IO.StreamWriter(_writerFileName);
                }
                catch (Exception e)
                {
                    _writer = null;
                    ReportFileOpenError(_writerFileName, "writing", e.Message);
                }

                return _writer;
            }
        }

        System.IO.TextReader _reader;
        public System.IO.TextReader Reader
        {
            get
            {
                // We can't read and write at the same time
                if (_writer != null)
                {
                    _data.ExecutionContext.ExecutionInformation.ReportError(String.Format("Trying to read from {0}, which is already open for writing.", _readerFileName ?? "[No file specified]"));
                    return null;
                }

                // If we don't already have a reader, create one
                try
                {
                    if (String.IsNullOrEmpty(_readerFileName))
                        ReportFileOpenError("[No file specified]", "reading", "[None]");
                    if (_reader == null)
                        _reader = new System.IO.StreamReader(_readerFileName);
                }
                catch (Exception e)
                {
                    _reader = null;
                    ReportFileOpenError(_readerFileName, "reading", e.Message);
                }

                return _reader;
            }
        }

        void ReportFileOpenError(string fileName, string action, string exceptionMessage)
        {
            _data.ExecutionContext.ExecutionInformation.ReportError(String.Format("Error opening {0} for {1}. This may mean the specified file, path or disk does not exist.\n\nInternal exception message: {2}", fileName, action, exceptionMessage));
        }

        #region IElement Members

        /// <summary>
        /// Method called when the simulation run is initialized.
        /// </summary>
        public void Initialize()
        { 
            // No initialization logic needed, we will open the file on the first read or write request
        }

        /// <summary>
        /// Method called when the simulation run is terminating.
        /// </summary>
        public void Shutdown()
        {
            // On shutdown, we need to make sure to close the file

            if (_writer != null)
            {
                try
                {
                    _writer.Close();
                    _writer.Dispose();
                }
                catch
                {
                }
                finally
                {
                    _writer = null;
                }
            }

            if (_reader != null)
            {
                try
                {
                    _reader.Close();
                    _reader.Dispose();
                }
                catch
                {
                }
                finally
                {
                    _reader = null;
                }
            }
        }

        #endregion
    }
}
